#!/usr/bin/env python3

"""
(C) 2019 Damir Cavar, Oren Baldinger, Maanvitha Gongalla, Anurag Kumar, Murali Kammili, Boli Fang

Wrappers for Xrenner to JSON-NLP output format.

Licensed under the Apache License 2.0, see the file LICENSE for more details.

Brought to you by the NLP-Lab.org (https://nlp-lab.org/)!
"""
import functools
from collections import OrderedDict, defaultdict
from os import getenv

from bs4 import BeautifulSoup
from dotenv import load_dotenv
from nltk import PunktSentenceTokenizer
from pyjsonnlp import build_coreference, find_head, remove_empty_fields
from pyjsonnlp.conversion import parse_conllu
from pyjsonnlp.pipeline import Pipeline
from nltk.parse.corenlp import CoreNLPDependencyParser
from xrenner import Xrenner

name = "xrennerjsonnlp"
__version__ = "0.0.1"
__cache = defaultdict(dict)

load_dotenv()


def cache_it(func):
    """A decorator to cache function response based on params. Add it to top of function as @cache_it."""

    global __cache

    @functools.wraps(func)
    def cached(*args):
        f_name = func.__name__
        s = ''.join(map(str, args))
        if s not in __cache[f_name]:
            __cache[f_name][s] = func(*args)
        return __cache[f_name][s]
    return cached


@cache_it
def load_xrenner():
    return Xrenner()


class CoreNLP:
    def __init__(self):
        self.parser = CoreNLPDependencyParser(url=self.corenlp_server())
        self.sentence_tokenizer = PunktSentenceTokenizer()

    @staticmethod
    def corenlp_server():
        return getenv('CORENLP_SERVER')

    def dep_parse(self, text: str, conll_version=10) -> str:
        """Get a CoreNLP depparse,lemma"""

        def get_conll(t):
            deps, = self.parser.raw_parse(t)
            return deps.to_conll(conll_version)  # xrenner requires conll10

        sentences = self.sentence_tokenizer.sentences_from_text(text)
        return '\n'.join(map(get_conll, sentences))


class XrennerPipeline(Pipeline):
    def __init__(self):
        self.corenlp = CoreNLP()

    def process(self, text='', coreferences=False, constituents=False, dependencies=False, expressions=False,
                lang='en') -> OrderedDict:

        x = load_xrenner()
        x.load(XrennerPipeline.iso2xrenner(lang))
        x.set_doc_name('not-used')  # needs to be set or error

        # do initial coreference parse
        conll = self.corenlp.dep_parse(text)
        sgml_result = x.analyze(conll, 'sgml')
        j = parse_conllu(conll)
        d = list(j['documents'].values())[0]

        d['meta']['DC.source'] = 'Xrenner 2.0'
        d['text'] = text

        # wrap tokens with their token id so that xml parsing works
        token_num = 1
        tokenized = []
        for line in sgml_result.split('\n'):
            if line[0:9] != '<referent' and line[0:10] != '</referent':
                line = f'<token id="{token_num}">{line}</token>'
                token_num += 1
            tokenized.append(line)

        representatives = {}
        coref_id = 0
        soup = BeautifulSoup('\n'.join(tokenized), 'html.parser')
        for tag in soup.find_all('referent'):
            # new representative
            if 'antecedent' not in tag.attrs or tag['type'] == 'none':
                r = build_coreference(coref_id)
                coref_id += 1
                r['representative'] = {
                    'entity': tag['entity'],
                    'tokens': [int(t['id']) for t in tag.find_all('token')]
                }
                r['representative']['head'] = find_head(d, r['representative']['tokens'])
                representatives[(tag['id'], tag['group'])] = r
                d['coreferences'].append(r)

                # might be a multi-word expression too!
                if expressions and tag['entity'] != 'event' and len(r['representative']['tokens']) > 1:
                    d['expressions'].append({
                        # deduce the phrase type by the pos tag of the head token
                        'type': 'VP' if 'V' in d['tokenList'][r['representative']['head']]['upos'] else 'NP',
                        'head': r['representative']['head'],
                        'tokens': r['representative']['tokens']
                    })
            # new referent
            else:
                r = representatives[(tag['antecedent'], tag['group'])]
                ids = [int(t['id']) for t in tag.find_all('token')]
                r['referents'].append({
                    'type': tag['type'],
                    'tokens': ids,
                    'head': find_head(d, ids)
                })

        return remove_empty_fields(j)

    @staticmethod
    def get_sentence_tokenizer():
        # https://textminingonline.com/dive-into-nltk-part-ii-sentence-tokenize-and-word-tokenize
        return PunktSentenceTokenizer()

    @staticmethod
    def iso2xrenner(lang: str) -> str:
        """get a three-letter date"""
        if lang == 'en':
            return 'eng'
        raise TypeError(f'Unsupported language {lang}')



from collections import OrderedDict
from unittest import TestCase, mock
from . import mocks

import pytest

from xrennerjsonnlp import XrennerPipeline


class TestXrenner(TestCase):
    @mock.patch('xrennerjsonnlp.CoreNLP.dep_parse')
    def test_process(self, dep_parse):
        dep_parse.return_value = """1	John	John	NNP	NNP	_	2	compound	_	_
2	Smith	Smith	NNP	NNP	_	3	nsubj	_	_
3	visited	visit	VBD	VBD	_	0	ROOT	_	_
4	Spain	Spain	NNP	NNP	_	3	dobj	_	_
5	.	.	.	.	_	3	punct	_	_

1	His	he	PRP$	PRP$	_	2	nmod:poss	_	_
2	visit	visit	NN	NN	_	3	nsubj	_	_
3	went	go	VBD	VBD	_	0	ROOT	_	_
4	well	well	RB	RB	_	3	advmod	_	_
5	.	.	.	.	_	3	punct	_	_
"""
        actual = XrennerPipeline().process("John Smith visited Spain. His visit went well.", 'en')
        expected = OrderedDict([('meta', OrderedDict([('DC.conformsTo', '0.2.4'), ('DC.created', '2019-01-25T17:04:34'), ('DC.date', '2019-01-25T17:04:34')])), ('documents', {'1': OrderedDict([('meta', OrderedDict([('DC.conformsTo', '0.2.4'), ('DC.source', 'Xrenner 2.0'), ('DC.created', '2019-01-25T17:04:34'), ('DC.date', '2019-01-25T17:04:34')])), ('id', '1'), ('text', 'John Smith visited Spain. His visit went well.'), ('tokenList', {1: {'id': 1, 'text': 'John', 'lemma': 'John', 'upos': 'NNP', 'xpos': 'NNP', 'features': OrderedDict([('Overt', 'Yes')])}, 2: {'id': 2, 'text': 'Smith', 'lemma': 'Smith', 'upos': 'NNP', 'xpos': 'NNP', 'features': OrderedDict([('Overt', 'Yes')])}, 3: {'id': 3, 'text': 'visited', 'lemma': 'visit', 'upos': 'VBD', 'xpos': 'VBD', 'features': OrderedDict([('Overt', 'Yes')])}, 4: {'id': 4, 'text': 'Spain', 'lemma': 'Spain', 'upos': 'NNP', 'xpos': 'NNP', 'features': OrderedDict([('Overt', 'Yes')])}, 5: {'id': 5, 'text': '.', 'lemma': '.', 'upos': '.', 'xpos': '.', 'features': OrderedDict([('Overt', 'Yes')])}, 6: {'id': 6, 'text': 'His', 'lemma': 'he', 'upos': 'PRP$', 'xpos': 'PRP$', 'features': OrderedDict([('Overt', 'Yes')])}, 7: {'id': 7, 'text': 'visit', 'lemma': 'visit', 'upos': 'NN', 'xpos': 'NN', 'features': OrderedDict([('Overt', 'Yes')])}, 8: {'id': 8, 'text': 'went', 'lemma': 'go', 'upos': 'VBD', 'xpos': 'VBD', 'features': OrderedDict([('Overt', 'Yes')])}, 9: {'id': 9, 'text': 'well', 'lemma': 'well', 'upos': 'RB', 'xpos': 'RB', 'features': OrderedDict([('Overt', 'Yes')])}, 10: {'id': 10, 'text': '.', 'lemma': '.', 'upos': '.', 'xpos': '.', 'features': OrderedDict([('Overt', 'Yes')])}}), ('sentences', {'0': {'id': '0', 'tokenFrom': 1, 'tokenTo': 6, 'tokens': [1, 2, 3, 4, 5]}, '1': {'id': '1', 'tokenFrom': 6, 'tokenTo': 11, 'tokens': [6, 7, 8, 9, 10]}}), ('dependencies', [{'style': 'universal', 'arcs': {1: [{'label': 'compound', 'governor': 2, 'dependent': 1}], 2: [{'label': 'nsubj', 'governor': 3, 'dependent': 2}], 3: [{'label': 'root', 'governor': 0, 'dependent': 3}], 4: [{'label': 'dobj', 'governor': 3, 'dependent': 4}], 5: [{'label': 'punct', 'governor': 3, 'dependent': 5}], 6: [{'label': 'nmod:poss', 'governor': 7, 'dependent': 6}], 7: [{'label': 'nsubj', 'governor': 8, 'dependent': 7}], 8: [{'label': 'root', 'governor': 0, 'dependent': 8}], 9: [{'label': 'advmod', 'governor': 8, 'dependent': 9}], 10: [{'label': 'punct', 'governor': 8, 'dependent': 10}]}}, {'style': 'enhanced', 'arcs': {}}]), ('coreferences', [{'id': 0, 'representative': {'entity': 'person', 'tokens': [1, 2], 'head': 2}, 'referents': [{'type': 'ana', 'tokens': [6], 'head': 6}]}, {'id': 1, 'representative': {'entity': 'event', 'tokens': [3], 'head': 3}, 'referents': [{'type': 'coref', 'tokens': [6, 7], 'head': 7}]}])])})])
        assert expected == actual, actual

    def test_iso2xrenner(self):
        actual = XrennerPipeline().iso2xrenner('en')
        assert 'eng' == actual, actual

        with pytest.raises(TypeError):
            XrennerPipeline().iso2xrenner('martian')